/*
 * Copyright (c) 2003 Gunnar Ritter
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute
 * it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source distribution.
 */
/*	Sccsid @(#)getdir.c	1.8 (gritter) 5/1/04	*/

#ifndef	__linux__
/*
 * 32-bit Solaris and Open UNIX do not have 64-bit getdents(); but
 * having _FILE_OFFSET_BITS=64 will make it use a dirent64 struct
 * on Open UNIX -> SEGV.
 */
#undef	_FILE_OFFSET_BITS
#endif	/* !__linux__ */

#include	<sys/types.h>
#include	<sys/stat.h>
#include	<stdlib.h>
#include	<errno.h>
#include	<string.h>

#if defined	(__GLIBC__) || defined (__FreeBSD__)
#include	<dirent.h>
#define	getdents(a, b, c)	getdirentries((a), (char *)(b), (c), &(db->g_offs))
#ifdef	__FreeBSD__
#undef	d_ino
#endif	/* __FreeBSD__ */
#elif defined	(__dietlibc__)
#include	<dirent.h>
#include	<unistd.h>
#else		/* !__GLIBC__, !__dietlibc__ */
#include	<sys/dirent.h>
#endif		/* !__GLIBC__, !__dietlibc__ */

#include	"getdir.h"

#define	DIBSIZE	5120

struct	getdb {
#ifndef	__FreeBSD__
	off_t		g_offs;
#else	/* __FreeBSD__ */
	long		g_offs;
#endif	/* __FreeBSD__ */
	struct dirent	*g_dirp;
	const char	*g_path;
	struct direc	g_dic;
	char		g_dirbuf[DIBSIZE+1];
	int		g_num;
	int		g_fd;
};

struct getdb *
getdb_alloc(const char *path, int fd)
{
	struct getdb	*db;

	if ((db = malloc(sizeof *db)) == NULL)
		return NULL;
	db->g_dirp = NULL;
	db->g_offs = 0;
	db->g_fd = fd;
	db->g_path = path;
	return db;
}

void
getdb_free(struct getdb *db)
{
	free(db);
}

struct direc *
getdir(struct getdb *db, int *err)
{
	*err = 0;
	while (db->g_dirp == NULL)
	{
		/*LINTED*/
		db->g_num = getdents(db->g_fd, (struct dirent *)db->g_dirbuf,
				DIBSIZE);
		if (db->g_num <= 0) {
			if (db->g_num < 0)
				*err = errno;
			db->g_offs = 0;
			return NULL;
		}
		/*LINTED*/
		db->g_dirp = (struct dirent *)db->g_dirbuf;
		while (db->g_dirp &&
#ifndef	__FreeBSD__
				db->g_dirp->d_ino == 0
#else	/* __FreeBSD__ */
				(db->g_dirp->d_fileno == 0 ||
				  db->g_dirp->d_type == DT_WHT)
#endif	/* __FreeBSD__ */
		      )
		{
		next:	if ((db->g_num -= db->g_dirp->d_reclen) == 0 ||
					db->g_dirp->d_reclen == 0)
				db->g_dirp = NULL;
			else
				db->g_dirp =
					/*LINTED*/
					(struct dirent *)((char *)db->g_dirp
						+ db->g_dirp->d_reclen);
		}
	}
#ifndef	__FreeBSD__
	if (db->g_dirp->d_ino == 0)
		goto next;
	db->g_dic.d_ino = db->g_dirp->d_ino;
#else	/* __FreeBSD__ */
	if (db->g_dirp->d_fileno == 0 || db->g_dirp->d_type == DT_WHT)
		goto next;
	db->g_dic.d_ino = db->g_dirp->d_fileno;
#endif	/* __FreeBSD__ */
	db->g_dic.d_name = db->g_dirp->d_name;
	if ((db->g_num -= db->g_dirp->d_reclen) == 0 ||
			db->g_dirp->d_reclen == 0)
		db->g_dirp = NULL;
	else
		/*LINTED*/
		db->g_dirp = (struct dirent *)((char *)db->g_dirp +
				db->g_dirp->d_reclen);
	return &(db->g_dic);
}
